#define _XOPEN_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include <mpi.h>

#define max(a,b) ({ __typeof__ (a) _a = (a); __typeof__ (b) _b = (b); _a > _b ? _a : _b; })
#define min(a,b) ({ __typeof__ (a) _a = (a); __typeof__ (b) _b = (b); _a < _b ? _a : _b; })

size_t offset(size_t total, size_t rank, size_t size)
{
	return rank * (total / size) + min(total % size, rank);
}
size_t length(size_t total, size_t rank, size_t size)
{
	return total / size + (total % size > rank);
}

// Swap two (double) elements
void swap(double *p0, double *p1)
{
	double v = *p0;
	*p0 = *p1;
	*p1 = v;
}

// Quicksort on double arrays : O(s log s)
void quicksort(double *a, size_t s)
{
	if (s)
	{
		size_t p = 0;
		for (size_t i=1; i<s; ++i)
			if (a[0] > a[i])
				swap(&a[i], &a[++p]);
		swap(&a[0], &a[p]);
		quicksort(&a[0  ], p    );
		quicksort(&a[p+1], s-p-1);
	}
}

// Merge sorted arrays : O(size1 + size2)
void merge_sorted(double *src1, size_t size1, double *src2, size_t size2, double *dst)
{
	size_t i1 = 0, i2 = 0;
	while (i1 < size1 && i2 < size2)
	{
		if (src1[i1] < src2[i2]) { dst[i1+i2] = src1[i1]; ++i1; }
		else                     { dst[i1+i2] = src2[i2]; ++i2; }
	}
	for (; i1<size1; ++i1) dst[i1+i2] = src1[i1];
	for (; i2<size2; ++i2) dst[i1+i2] = src2[i2];
}

int main(int argc, char *argv[])
{
	MPI_Init(&argc, &argv);

	// Default size = 17
	size_t total = 17ul;
	if (argc > 1)
	{
		int _t = atoi(argv[1]);
		if (_t < 0)
		{
			fprintf(stderr, "[INVALID ARG] Size must be non negative !\n");
			MPI_Finalize();
			return 0;
		}
		total = _t;
	}

	// Get world characteristics
	int rank, size;
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &size);

	// Initialize random seed
	srand48(time(NULL) + rank);

	// Local Table
	size_t  As = length(total, rank, (size_t) size);
	double *Ai = calloc(As, sizeof(double));

	// Random Initialization of local Ai
	for (size_t i = 0ul; i<As; ++i) Ai[i] = drand48();

	MPI_Barrier(MPI_COMM_WORLD);
	double t1 = MPI_Wtime();
	
	// Quicksort the initial local block : O(s log s)
	quicksort(Ai, As);
	
	/***************************************************************************
	 *                        The magic happens here !                         *
	 ***************************************************************************/
	for (size_t window = 1ul; window < (size_t) size; window *= 2)
		if (rank % (2 * window) == window) // Send case
		{
			// We send our whole block to rank - window
			MPI_Send(Ai, As, MPI_DOUBLE, rank - window, 0, MPI_COMM_WORLD);
			free(Ai);
			As = 0;
		}
		else if (rank % (2 * window) == 0 && rank + window < (size_t) size) // Receive case
		{
			// We receive from rank + window
			size_t recv_rank  = rank + window;
			// A block that contains everything up to min(rank + 2 window, size)
			size_t recv_last  = min(recv_rank + window, (size_t) size);
			// Offset of the start of the received block
			size_t recv_begin = offset(total, recv_rank, (size_t) size);
			// Offset of the end of the received block
			size_t recv_end   = offset(total, recv_last, (size_t) size);

			// Allocate buffer for receiving
			size_t  Bs = recv_end - recv_begin;
			double *Bi = calloc(Bs, sizeof(double));

			// Receive block
			MPI_Recv(Bi, Bs, MPI_DOUBLE, recv_rank, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);

			// Allocate memory space to merge
			size_t  Cs = As + Bs;
			double *Ci = calloc(Cs, sizeof(double));

			// Merge
			merge_sorted(Ai, As, Bi, Bs, Ci);

			// Cleanup + update
			free(Ai);
			free(Bi);
			Ai = Ci;
			As = Cs;
		}

	// Rank 0 has everything and print the sorted table
	/*
	if (rank == 0)
		for (size_t i = 0ul; i<As; ++i)
			printf("[%.3lu] %f\n", i, Ai[i]);
	*/

	MPI_Barrier(MPI_COMM_WORLD);
	double t2 = MPI_Wtime();
	if (rank==0) printf("%d %lu %f\n", size, total, t2-t1);

	MPI_Finalize();

	return 0;
}
