#define _XOPEN_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <mpi.h>

#define max(a,b) ({ __typeof__ (a) _a = (a); __typeof__ (b) _b = (b); _a > _b ? _a : _b; })
#define min(a,b) ({ __typeof__ (a) _a = (a); __typeof__ (b) _b = (b); _a < _b ? _a : _b; })

size_t offset(size_t total, size_t rank, size_t size)
{
	return rank * (total / size) + min(total % size, rank);
}
size_t length(size_t total, size_t rank, size_t size)
{
	return total / size + (total % size > rank);
}

// Swap two (double) elements
void swap(double *p0, double *p1)
{
	double v = *p0;
	*p0 = *p1;
	*p1 = v;
}

// Quicksort on double arrays : O(s log s)
void quicksort(double *a, size_t s)
{
	if (s)
	{
		size_t p = 0;
		for (size_t i=1; i<s; ++i)
			if (a[0] > a[i])
				swap(&a[i], &a[++p]);
		swap(&a[0], &a[p]);
		quicksort(&a[0  ], p    );
		quicksort(&a[p+1], s-p-1);
	}
}

// Merge sorted arrays : O(size1 + size2)
void merge_sorted(double *src1, size_t size1, double *src2, size_t size2, double *dst)
{
	size_t i1 = 0, i2 = 0;
	while (i1 < size1 && i2 < size2)
	{
		if (src1[i1] < src2[i2]) { dst[i1+i2] = src1[i1]; ++i1; }
		else                     { dst[i1+i2] = src2[i2]; ++i2; }
	}
	for (; i1<size1; ++i1) dst[i1+i2] = src1[i1];
	for (; i2<size2; ++i2) dst[i1+i2] = src2[i2];
}

int main(int argc, char *argv[])
{
	MPI_Init(&argc, &argv);

	// Default size = 17
	size_t total = 17ul;
	if (argc > 1)
	{
		int _t = atoi(argv[1]);
		if (_t < 0)
		{
			fprintf(stderr, "[INVALID ARG] Size must be non negative !\n");
			MPI_Finalize();
			return 0;
		}
		total = _t;
	}

	// Get world characteristics
	int rank, size;
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &size);

	// Initialize random seed
	srand48(time(NULL) + rank);

	// Local Table
	size_t  As = length(total, rank, (size_t) size);
	double *Ai = calloc(As, sizeof(double));

	// Random Initialization of local Ai
	for (size_t i = 0ul; i<As; ++i) Ai[i] = drand48();

	MPI_Barrier(MPI_COMM_WORLD);
	double t1 = MPI_Wtime();

	// Quicksort the initial local block : O(s log s)
	quicksort(Ai, As);

	/***************************************************************************
	 *                        The magic happens here !                         *
	 ***************************************************************************/

	 // Creating temporary buffers
	 size_t  Bs = length(total, rank+1, (size_t) size);
	 double *Bi = calloc(Bs, sizeof(double));
	 size_t  Cs = As + Bs;
	 double *Ci = calloc(Cs, sizeof(double));

	for (int step = 0; step < size; ++step)
		if (rank % 2 == step % 2 && rank + 1 < size)
		{
			// printf("[%d] %d <= %d\n", step, rank, rank+1);
			MPI_Recv(Bi, Bs, MPI_DOUBLE, rank + 1, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
			merge_sorted(Ai, As, Bi, Bs, Ci);
			MPI_Send(&Ci[As], Bs, MPI_DOUBLE, rank + 1, 0, MPI_COMM_WORLD);
			memcpy(Ai, Ci, As * sizeof(double));
		}
		else if (rank % 2 != step % 2 && rank > 0)
		{
			// printf("[%d] %d <= %d\n", step, rank-1, rank);
			MPI_Send(Ai, As, MPI_DOUBLE, rank - 1, 0, MPI_COMM_WORLD);
			MPI_Recv(Ai, As, MPI_DOUBLE, rank - 1, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
		}

	// Cleaning temporary buffers
	free(Bi);
	free(Ci);

	// Printing the distributed array
	/*
	for (int r = 0; r < size; ++r)
	{
		if (r == rank)
			for (size_t i = 0ul; i<As; ++i)
				printf("[%.2d|%.3lu] %f\n", r, i, Ai[i]);
		MPI_Barrier(MPI_COMM_WORLD);
	}
	*/
	
	MPI_Barrier(MPI_COMM_WORLD);
	double t2 = MPI_Wtime();
	if (rank==0) printf("%d %lu %f\n", size, total, t2-t1);

	MPI_Finalize();

	return 0;
}
